/* FILE: evoc.h                    -*-Mode: c++-*-
 *
 * Stuff that should be moved eventually into the oc extension.
 * 
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2003/11/20 22:41:30 $
 * Last modified by: $Author: dgp $
 */

#ifndef _NB_EVOC
#define _NB_EVOC

#include "oc.h"

#include <stdio.h>
#include <math.h>
#include <time.h>

/* End includes */     /* Optional directive to build.tcl */

#ifndef PI
#define PI             3.14159265358979323846
#endif

#ifndef SQRT2
#define SQRT2          1.41421356237309504880
#endif

#ifndef SQRT1_2
#define	SQRT1_2        0.70710678118654752440      /* 1/sqrt(2) */
#endif

#ifndef MU0
#define MU0           12.56637061435917295385e-7   /* 4 PI 10^7 */
#endif

#define SCRATCH_BUF_SIZE 1024  // Size of scratch space buffers
#define ASCIIPTR_BUF_SIZE (POINTERASCIIWIDTH+16)
/// Room + extra for %p representation of pointers


extern int Verbosity;
#ifndef DEBUGLVL
#define DEBUGLVL 0  // 0 => No debug code, 100 => Full debug code
#endif 

// Several useful macros
#define OC_SQ(x) ((x)*(x))
#define OC_MIN(x,y) ((x)<(y)?(x):(y))
#define OC_MAX(x,y) ((x)>(y)?(x):(y))
#define OC_ROUND(x) ((x)<0 ? ceil((x)-0.5) : floor((x)+0.5))
/// Note: rint() is not in the ANSI C spec, and in any event
///  may vary depending on "prevalent rounding mode".  Be
///  aware, however, that the above has a bias towards the
///  larger magnitude integer (which could be removed by
///  rounding x.5 to the even integer).


//////////////////////////////////////////////////////////////////////////
// Class documentation structure
struct ClassDoc {
public:
  const CHAR *classname;
  const CHAR *maintainer;
  const CHAR *revision;
  const CHAR *revdate;
  ClassDoc(const CHAR *new_classname,const CHAR *new_maintainer,
	   const CHAR *new_revision,const CHAR *new_revdate);
};

//////////////////////////////////////////////////////////////////////////
// Tcl/Tk Interface

// C-pointer <--> ascii string conversion class.
//   Currently this routine maps things through %x or %lx, but this
// may change in the future to use a hash table.  The client should
// first call GetAsciiSize to determine how large a buffer is needed
// to hold the ascii-fied pointer string, then call PtrToAscii to
// get a string representation of the ptr.  This can be passed across
// to Tcl procs as needed.  Later, when a C routine is called by a
// Tcl proc having such a string, this string can be converted back
// to a valid ptr via AsciiToPtr.  NOTE: AsciiToPtr should ONLY be
// called with strings produced by PtrToAscii.  There is currently
// no check on this, but there will be if this class gets modified
// to use a hash table lookup scheme.
//
class Omf_AsciiPtr
{
private:
  static const ClassDoc class_doc;
  static const size_t ascii_string_width;
public:
  static size_t GetAsciiSize() { return ascii_string_width; }

  static void PtrToAscii(const void* ptr,char* buf);
  /// It is the responsibility of the calling routine to insure
  /// that buf is dimensioned to at least GetStringSize() bytes.

  static void* AsciiToPtr(const char* buf);
  /// NOTE 1: ONLY strings produced by Omf_AsciiPtr::PtrToAscii
  ///  should ever be sent to this routine. 
  /// Note 2: For consistency, I would prefer to make this function
  ///    static void AsciiToPtr(const char* buf,void * &ptr)
  ///  but gcc 2.7.2.1 (at least) screws up implicit casting
  ///  to void *&, apparently producing a temporary which never
  ///  gets written back into ptr.  So the above assignment seems
  ///  a safer option.  (The alternative is to be _real_ sure
  ///  that the client routine does an explicit cast to void *&.)
};

//////////////////////////////////////////////////////////////////////////
// Utility routines

// Random number generator
void Oc_Srand();
void Oc_Srand(unsigned int seed);
double Oc_UnifRand();
Tcl_CmdProc OcSrand;
Tcl_CmdProc OcUnifRand;
/// Oc_UnifRand() random value in [0,1] with unif. distrib.
/// The random number generator may be (re)initialized by
/// calling Oc_Srand() or Oc_Srand(seed).  In the first case
/// the seed is determined by sampling the system clock.

// Extra goodies for Tcl's expr command.  The Tcl_CmdProc version
// is intended for use at the script level when setting up slave
// interpreters.  The main interpreter has this done for it during
// Nb_Init().
void Oc_AddTclExprExtensions(Tcl_Interp* interp);
Tcl_CmdProc OcAddTclExprExtensions;


//////////////////////////////////////////////////////////////////////////
// Oc_Times function
//   This is a portable replacement for the Unix times(2) function
class Oc_TimeVal; // Forward declaration
void Oc_Times(Oc_TimeVal& cpu_time,Oc_TimeVal& wall_time);
Tcl_CmdProc OcTimes;

//////////////////////////////////////////////////////////////////////////
// Oc_TimeVal class
//  NOTE: This class supports only non-negative times.
class Oc_TimeVal {
  friend void Oc_Times(Oc_TimeVal& cpu_time,Oc_TimeVal& wall_time);
private:
  static const ClassDoc class_doc;
  unsigned long ticks_per_second;
  unsigned long max_ticks;  // Largest allowed value for ticks
  /// The total number of ticks is overflow*(max_ticks+1)+ticks.
  unsigned long ticks;
  unsigned long overflow;
  void Print(FILE* fptr);  // For debugging
public:
  void Reset();
  void Reset(unsigned long _ticks_per_second,unsigned long _max_ticks);

  Oc_TimeVal();
  Oc_TimeVal(const Oc_TimeVal& time);
  Oc_TimeVal(unsigned long _ticks_per_second,unsigned long _max_ticks);

  unsigned long GetTicksPerSecond() const { return ticks_per_second; }
  unsigned long GetMaxTicks() const { return max_ticks; }
  unsigned long GetTicks() const { return ticks; }
  void SetTicks(unsigned long _ticks);
  unsigned long GetOverflow() const { return overflow; }
  void SetOverflow(unsigned long _overflow) { overflow=_overflow; }
  BOOL IsValid() const;
  operator double() const; // Returns time in seconds, in floating point
  Oc_TimeVal& operator=(const Oc_TimeVal& time);
  Oc_TimeVal& operator+=(const Oc_TimeVal& time);
  Oc_TimeVal& operator-=(const Oc_TimeVal& time);
  friend BOOL AreCompatible(const Oc_TimeVal& time1,
			    const Oc_TimeVal& time2);
  friend Oc_TimeVal operator+(const Oc_TimeVal& time1,
			      const Oc_TimeVal& time2);
  friend Oc_TimeVal operator-(const Oc_TimeVal& time1,
			      const Oc_TimeVal& time2);
};

#endif // _NB_EVOC
